/*
 *  audio_player.c 
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  ========================================================================
 *
 *  Assume deal with raw audio data formatted as
 * 	Signed 16bit, Little Endian
 *	Stereo
 *	32k Hz
 *
 *  To convert raw PCM data to Wav file
 *	[prompt]# sox -V -s -w -c 2 -r 32000
 *		ps. a 44B header is attached
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/soundcard.h>
#include <asm/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include "videodev2.h"

#define	BUF_SIZE	4096
#define MAX_BUFFER	64
#define SampleRate	32000
#define CTRL_DEV	"/dev/video0"
#define SOURCE_DEV	"/dev/video2"
#define SOUND_DEV	"/dev/dsp"

int sound_fd=-1, source_fd=-1, ctrl_fd=-1;
int doabort=0;
unsigned char audio_buffer[BUF_SIZE];
char *output_ptrs[MAX_BUFFER];
char *gVAMap;
int gMapSize;

void usage()
{
        printf("Usage: audio_player [OPTIONS]\n\n");
        printf("OPTIONS\n");
        printf("  -d DSP_Device : The DSP device you want to manipulate.\n");
	printf("                  For sound output, usually a soundcard.\n");
        printf("                  Default \"/dev/dsp\"\n");
        printf("\n");
        exit(1);
}

void set_mute(int fd, int mute)
{
	struct v4l2_control ctrl;

	/* VIDIOC_S_CTRL */
	ctrl.id = V4L2_CID_AUDIO_MUTE;
	ctrl.value = mute;
	if( ioctl(fd, VIDIOC_S_CTRL, &ctrl) )
	 	perror("VIDIOC_S_CTRL");
}

int allocate_buffer(int *dev_off, int *count, int *buf_sz)
{
	struct v4l2_requestbuffers reqbuf;
	struct v4l2_buffer buf;
	int offset = -1, size;
	int ret, i;

	// request buffer
	reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	reqbuf.memory = V4L2_MEMORY_MMAP;
	reqbuf.count = *count;
	ret = ioctl(source_fd, VIDIOC_REQBUFS, &reqbuf);
	if(ret) {
		perror("VIDIOC_REQBUFS");
		return 0;
	}

	for(i=0, size=0, *count=0; i<reqbuf.count; ++i) {
		buf.type = reqbuf.type;
		buf.index = i;
		ret = ioctl(source_fd, VIDIOC_QUERYBUF, &buf);
		
		
		if(ret) {
			perror("VIDIOC_QUERYBUF");
			return 0;
		}
		*buf_sz = buf.length;
		
		if(0==size) offset = *dev_off = buf.m.offset;
		if(offset==buf.m.offset) {
			offset += buf.length;
			size += buf.length;
			(*count)++;
		}
	} // end of for loop
	
	*count = reqbuf.count;

	return size;
}

int do_map_n_q(void)
{
	struct v4l2_buffer buf;
	int dev_start, buf_size, buf_count=MAX_BUFFER;
	char *map_start;
	int i;

	gMapSize = allocate_buffer(&dev_start, &buf_count, &buf_size);
	if(0==gMapSize) {
		printf("zero size to map\n");
		return 0;
	}

	gVAMap = mmap(0, gMapSize, PROT_READ|PROT_WRITE, MAP_SHARED, 
			source_fd, dev_start);

	if(MAP_FAILED==gVAMap) {
		perror("mmap");
		return 0;
	}

	for(i=0, map_start=gVAMap; i<buf_count; ++i, map_start+=buf_size) {
		memset(&buf, 0, sizeof(buf));
		buf.index = i;
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		if(ioctl(source_fd, VIDIOC_QBUF, &buf)) {
			perror("VIDIOC_QBUF");
			munmap(gVAMap, gMapSize);
			return 0;
		}
		output_ptrs[i] = map_start;
	}

	return buf_count;

}

int play_streaming()
{
	unsigned int readcount=0;
	int buf_count;
	int i, j;
	struct v4l2_buffer buf;
	int ret = 0, error=0;

	buf_count = do_map_n_q();
	if(0==buf_count) return 0;

	if(ioctl(source_fd, VIDIOC_STREAMON, &i)) goto done;

	i=1;
	while(!doabort) {
		for(j=0; j<buf_count; ++j) {

			if(doabort)
			break;
		
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			if((ret=ioctl(source_fd, VIDIOC_DQBUF, &buf))) {
				error = errno;
				goto err;
			}

			// write to dsp
			ret = write(sound_fd, output_ptrs[j],
				    buf.bytesused);

			if(-1 == ret) {
				perror("write");
				goto err;
			}
			else if(buf.bytesused != ret) {
				printf("cannot write full buffer\n");
			}
			
			if(ioctl(source_fd, VIDIOC_QBUF, &buf)) {
				perror("VIDIOC_QBUF");
				goto err;
			}

			readcount++;
		} // end of for loop
	} // end of while loop
	
err:
	if(ioctl(source_fd, VIDIOC_STREAMOFF, &i))
		perror("VIDIOC_STREAMOFF");
done:
	munmap(gVAMap, gMapSize);
	return error;
}

void sighandler(int signo)
{
	doabort = 1;
}

void print_oss_fmt(int fmt) {
	switch(fmt) {
	case AFMT_U8:
		printf("unsigned 8bit\n"); break;
	case AFMT_S16_LE:
		printf("signed 16bit little endian\n"); break;
	case AFMT_S16_BE:
		printf("signed 16bit big endian\n"); break;
	case AFMT_S8:
		printf("signed 8bit\n"); break;
	case AFMT_U16_LE:
		printf("unsigned 16bit little endian\n"); break;
	case AFMT_U16_BE:
		printf("unsigned 16bit big endian\n"); break;
	default:
		printf("0x%X\n", fmt); break;
	}
}

int main(int argc, char *argv[])
{
	int ret, format=AFMT_S16_LE;
	int channel=2, rate = SampleRate;
	char sound_dev[256] = SOUND_DEV;

	while(1) {
		int arg;

		if((arg=getopt(argc, argv, "hd:"))<0)
			break;

		switch(arg) {
		case '?':
		case 'h':
			usage();
			break;
		case 'd':
			strncpy(sound_dev, optarg, sizeof(sound_dev));
			break;
		}
	}

	// allow abort via Ctrl+C
	signal(SIGINT, sighandler);

	ctrl_fd = open(CTRL_DEV, O_RDWR);
	if(ctrl_fd == -1)
		printf("cannot open ctrl-dev \"%s\": %s\n", CTRL_DEV, strerror(errno));

	// sound output fd
	sound_fd = open(sound_dev, O_RDWR);
	printf("open soundcard dsp %s ", sound_dev);

	if(sound_fd < 0) {
		printf("failed: %s\n", strerror(errno));
		exit(errno);
	}
	else {
		printf("done\n");	
	}	

	printf("open source dev %s ", SOURCE_DEV);
	source_fd = open(SOURCE_DEV, O_RDWR);
	if( -1==source_fd ) {
		printf("failed: %s\n", strerror(errno));
		exit(errno);
	}
	else {
		printf("done\n");
	}

	if( (ret=ioctl(sound_fd, SNDCTL_DSP_SETFMT, &format)) ){
		printf("SNDCTL_DSP_SETFMT: %s\n", strerror(errno));
		return -1;
	}
	else {
		printf("accepted oss format: ");
		print_oss_fmt(format);
	}

	if( (ret=ioctl(sound_fd, SNDCTL_DSP_CHANNELS, &channel)) ){
		printf("SNDCTL_DSP_CHANNELS: %s\n", strerror(errno));
		return -1;
	}
	else
		printf("accepted dsp channels: %d\n", channel);

	if( (ret=ioctl(sound_fd, SNDCTL_DSP_SPEED, &rate)) ){
		printf("SNDCTL_DSP_SPEED: %s\n", strerror(errno));
		return -1;
	}
	else
		printf("accepted sample rate: %d\n", rate);

	if( (ret=ioctl(sound_fd, SNDCTL_DSP_RESET, 0)) ){
		printf("SNDCTL_DSP_RESET: %s\n", strerror(errno));
		return -1;
	}

	set_mute(ctrl_fd, 0);
	return play_streaming();
}
